#!/usr/bin/env python3

import os
import os.path
import stanza
from stanza.utils.conll import CoNLL
import sys
from tqdm import tqdm

def oneLine(sentence):
    return sentence.text.replace('\n', ' ').replace('\t', ' ')

def coNLLGet(token, attribute):
    return str(token[attribute]) if attribute in token else '_'

def singleOrRange(token_id):
    if type(token_id) == int:
        return str(token_id)
    else:
        return '-'.join(map(str, token_id))

def miscAttribute(token):
    return lambda key: [f"{key}={token[key]}"] if key in token else []

def misc(token):
    keys = ['start_char', 'end_char', 'ner']
    attributes = sum(map(miscAttribute(token), keys), [])
    return '|'.join(attributes) if len(attributes) > 0 else '_'

def formatToken(token):
    return '\t'.join([
            singleOrRange(token['id']),
            coNLLGet(token, "text"),
            coNLLGet(token, "lemma"),
            coNLLGet(token, "upos"),
            coNLLGet(token, "xpos"),
            coNLLGet(token, "feats"),
            coNLLGet(token, "head"),
            coNLLGet(token, "deprel"),
            coNLLGet(token, "deps"),
            misc(token)
        ])

class Annotator:
    def __init__(self, source, target):
        self.source = source
        self.target = target
        self.paragraphSerial = -1
        self.sentenceSerial = -1
        self.model = stanza.Pipeline(
                lang='fr',
                processors='tokenize,pos,lemma,depparse,ner')

    def run(self):
        for name in os.listdir(self.source):
            if os.path.isdir(f'{self.source}/{name}'):
                self.annotate_dir(name)

    def annotate_dir(self, directory):
        source_path = f'{self.source}/{directory}'
        os.makedirs(f'{self.target}/{directory}', exist_ok=True)
        for name in tqdm(os.listdir(source_path)):
            path = f'{source_path}/{name}'
            relpath = f'{directory}/{name}'
            if os.path.isdir(path):
                self.annotate_dir(relpath)
            elif os.path.isfile(path):
                self.annotate_file(relpath)

    def annotate_file(self, file):
        outputFile = f'{self.target}/{file[:-4]}.conllu'
        if not os.path.isfile(outputFile):
            with open(f'{self.source}/{file}', 'r') as source:
                document = source.read()
                try:
                    with open(outputFile, 'w') as target:
                        for paragraph in document.split('\n\n'):
                            self.annotate_paragraph(paragraph, target)
                except Exception:
                    print("Failed trying " + outputFile)
                    os.unlink(outputFile)
                    sys.exit(1)

    def annotate_paragraph(self, paragraph, target):
        parsed = self.model(paragraph)
        if len(parsed.sentences) > 0:
            print(f'# newpar id = {self.newpar()}', file=target)
            for sentence in parsed.sentences:
                print(self.annotate_sentence(sentence), file=target)

    def annotate_sentence(self, sentence):
        return "\n".join(
                [f'# sent_id = {self.newsent()}',
                 f'# text = {oneLine(sentence)}'] +
                list(map(formatToken, sentence.to_dict())) +
                [''])

    def newpar(self):
        self.paragraphSerial += 1
        return self.paragraphSerial

    def newsent(self):
        self.sentenceSerial += 1
        return self.sentenceSerial

if __name__ == '__main__':
    Annotator(sys.argv[1], sys.argv[2]).run()
